/* Copyright (C) |Meso|Star> 2016-2017 (contact@meso-star.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>. */

#include "ssf.h"
#include "ssf_bxdf_c.h"
#include "ssf_optics.h"

#include <rsys/double3.h>

struct specular_transmission {
  struct ssf_fresnel* fresnel;
  double eta_i, eta_t;
};

/*******************************************************************************
 * Private functions
 ******************************************************************************/
static res_T
specular_transmission_init
  (struct mem_allocator* allocator, void* bxdf)
{
  struct specular_transmission* btdf = bxdf;
  ASSERT(btdf);
  return ssf_fresnel_create
    (allocator, &ssf_fresnel_dielectric_dielectric, &btdf->fresnel);
}

static void
specular_transmission_release(void* bxdf)
{
  struct specular_transmission* btdf = bxdf;
  ASSERT(btdf);
  if(btdf->fresnel) SSF(fresnel_ref_put(btdf->fresnel));
}

static double
specular_transmission_sample
  (void* bxdf,
   struct ssp_rng* rng,
   const double wo[3],
   const double N[3],
   double wi[3],
   int* type,
   double* pdf)
{
  struct specular_transmission* btdf = bxdf;
  double eta;
  double cos_wi_N;
  ASSERT(rng && wo && N && wi && type && pdf);
  ASSERT(d3_is_normalized(wo) && d3_is_normalized(N) && d3_dot(wo, N) > 0);
  (void)rng;

  SSF(fresnel_dielectric_dielectric_setup
    (btdf->fresnel, btdf->eta_t, btdf->eta_i));

  eta = btdf->eta_i / btdf->eta_t;
  if(!refract(wi, wo, N, eta)) { /* Total internal reflection */
    reflect(wi, wo, N);
    *pdf = INF;
    *type = SSF_REFLECTION | SSF_SPECULAR;
    return 0.0;
  }
  cos_wi_N = d3_dot(wi, N);
  *pdf = INF;
  *type = SSF_TRANSMISSION | SSF_SPECULAR;
  return 1.0 - ssf_fresnel_eval(btdf->fresnel, fabs(cos_wi_N));
}

static double
specular_transmission_eval
  (void* data, const double wo[3], const double N[3], const double wi[3])
{
  (void)data, (void)wi, (void)N, (void)wo;
  return 0.0;
}

static double
specular_transmission_pdf
  (void* data, const double wo[3], const double N[3], const double wi[3])
{
  (void)data, (void)wi, (void)N, (void)wo;
  return 0.0;
}

/*******************************************************************************
 * Exported symbols
 ******************************************************************************/
const struct ssf_bxdf_type ssf_specular_transmission = {
  specular_transmission_init,
  specular_transmission_release,
  specular_transmission_sample,
  specular_transmission_eval,
  specular_transmission_pdf,
  sizeof(struct specular_transmission),
  ALIGNOF(struct specular_transmission)
};

res_T
ssf_specular_transmission_setup
  (struct ssf_bxdf* bxdf, const double eta_i, const double eta_t)
{
  struct specular_transmission* btdf;
  if(!bxdf || eta_i <= 0 || eta_t <= 0) return RES_BAD_ARG;
  if(!BXDF_TYPE_EQ(&bxdf->type, &ssf_specular_transmission)) return RES_BAD_ARG;
  btdf = bxdf->data;
  btdf->eta_i = eta_i;
  btdf->eta_t = eta_t;
  return RES_OK;
}

